.sage-ai-unified-diff-active.has-merge-chunks .jp-Toolbar {
    display: none !important;
  }

.ͼ2.cm-merge-b .cm-changedText {
    background: none !important;
}
.code-mirror-empty-original-content .cm-deletedLine:first-of-type,
.code-mirror-empty-original-content .cm-deletedChunk:first-of-type > .cm-deletedLine {
  display: none !important;
}

.cm-collapsedLines {
  background: transparent !important;
}